@@ -152,6 +152,9 @@ static gint                  g_key_file_parse_value_as_integer (GKeyFile
 								GError                **error);
 static gchar                *g_key_file_parse_integer_as_value (GKeyFile               *key_file,
 								gint                    value);
+static gdouble               g_key_file_parse_value_as_double  (GKeyFile               *key_file,
+                                                                const gchar            *value,
+                                                                GError                **error);
 static gboolean              g_key_file_parse_value_as_boolean (GKeyFile               *key_file,
 								const gchar            *value,
 								GError                **error);
@@ -185,12 +188,7 @@ static void                  g_key_file_flush_parse_buffer     (GKeyFile
 GQuark
 g_key_file_error_quark (void)
 {
-  static GQuark error_quark = 0;
-
-  if (error_quark == 0)
-    error_quark = g_quark_from_static_string ("g-key-file-error-quark");
-
-  return error_quark;
+  return g_quark_from_static_string ("g-key-file-error-quark");
 }
 
 static void
@@ -355,7 +353,7 @@ g_key_file_load_from_fd (GKeyFile       *key_file,
 			 GError        **error)
 {
   GError *key_file_error = NULL;
-  gsize bytes_read;
+  gssize bytes_read;
   struct stat stat_buf;
   gchar read_buf[4096];
   
@@ -587,8 +585,11 @@ g_key_file_load_from_data_dirs (GKeyFile       *key_file,
 
   found_file = FALSE;
   data_dirs = all_data_dirs;
+  output_path = NULL;
   while (*data_dirs != NULL && !found_file)
     {
+      g_free (output_path);
+
       fd = find_file_in_data_dirs (file, &output_path, &data_dirs, 
                                    &key_file_error);
       
@@ -596,7 +597,7 @@ g_key_file_load_from_data_dirs (GKeyFile       *key_file,
         {
           if (key_file_error)
             g_propagate_error (error, key_file_error);
-	  break;
+ 	  break;
         }
 
       found_file = g_key_file_load_from_fd (key_file, fd, flags,
@@ -606,15 +607,17 @@ g_key_file_load_from_data_dirs (GKeyFile       *key_file,
       if (key_file_error)
         {
 	  g_propagate_error (error, key_file_error);
-          g_free (output_path);
 	  break;
         }
-      
-      if (full_path)
-	*full_path = output_path;
     }
 
+  if (found_file && full_path)
+    *full_path = output_path;
+  else 
+    g_free (output_path);
+
   g_strfreev (all_data_dirs);
+
   return found_file;
 }
 
@@ -1014,20 +1017,33 @@ g_key_file_get_keys (GKeyFile     *key_file,
       return NULL;
     }
 
-  num_keys = g_list_length (group->key_value_pairs);
-  
-  keys = (gchar **) g_new0 (gchar **, num_keys + 1);
-
-  tmp = group->key_value_pairs;
-  for (i = 1; i <= num_keys; i++)
+  num_keys = 0;
+  for (tmp = group->key_value_pairs; tmp; tmp = tmp->next)
     {
       GKeyFileKeyValuePair *pair;
 
       pair = (GKeyFileKeyValuePair *) tmp->data;
-      keys[num_keys - i] = g_strdup (pair->key);
 
-      tmp = tmp->next;
+      if (pair->key)
+	num_keys++;
     }
+  
+  keys = g_new0 (gchar *, num_keys + 1);
+
+  i = num_keys - 1;
+  for (tmp = group->key_value_pairs; tmp; tmp = tmp->next)
+    {
+      GKeyFileKeyValuePair *pair;
+
+      pair = (GKeyFileKeyValuePair *) tmp->data;
+
+      if (pair->key)
+	{
+	  keys[i] = g_strdup (pair->key);
+	  i--;
+	}
+    }
+
   keys[num_keys] = NULL;
 
   if (length)
@@ -1089,7 +1105,7 @@ g_key_file_get_groups (GKeyFile *key_file,
    * list) is always the comment group at the top,
    * which we skip
    */
-  groups = (gchar **) g_new0 (gchar **, num_groups);
+  groups = g_new0 (gchar *, num_groups);
 
   group_node = g_list_last (key_file->groups);
   
@@ -2140,6 +2156,218 @@ g_key_file_set_integer_list (GKeyFile     *key_file,
   g_string_free (values, TRUE);
 }
 
+/**
+ * g_key_file_get_double:
+ * @key_file: a #GKeyFile
+ * @group_name: a group name
+ * @key: a key
+ * @error: return location for a #GError
+ *
+ * Returns the value associated with @key under @group_name as an
+ * integer. If @group_name is %NULL, the start_group is used.
+ *
+ * If @key cannot be found then the return value is undefined and
+ * @error is set to #G_KEY_FILE_ERROR_KEY_NOT_FOUND. Likewise, if
+ * the value associated with @key cannot be interpreted as a double
+ * then the return value is also undefined and @error is set to
+ * #G_KEY_FILE_ERROR_INVALID_VALUE.
+ *
+ * Return value: the value associated with the key as a double.
+ *
+ * Since: 2.12
+ **/
+gdouble
+g_key_file_get_double  (GKeyFile     *key_file,
+                        const gchar  *group_name,
+                        const gchar  *key,
+                        GError      **error)
+{
+  GError *key_file_error;
+  gchar *value;
+  gdouble double_value;
+
+  g_return_val_if_fail (key_file != NULL, -1);
+  g_return_val_if_fail (group_name != NULL, -1);
+  g_return_val_if_fail (key != NULL, -1);
+
+  key_file_error = NULL;
+
+  value = g_key_file_get_value (key_file, group_name, key, &key_file_error);
+
+  if (key_file_error)
+    {
+      g_propagate_error (error, key_file_error);
+      return 0;
+    }
+
+  double_value = g_key_file_parse_value_as_double (key_file, value,
+                                                  &key_file_error);
+  g_free (value);
+
+  if (key_file_error)
+    {
+      if (g_error_matches (key_file_error,
+                           G_KEY_FILE_ERROR,
+                           G_KEY_FILE_ERROR_INVALID_VALUE))
+        {
+          g_set_error (error, G_KEY_FILE_ERROR,
+                       G_KEY_FILE_ERROR_INVALID_VALUE,
+                       _("Key file contains key '%s' in group '%s' "
+                         "which has value that cannot be interpreted."), key,
+                       group_name);
+          g_error_free (key_file_error);
+        }
+      else
+        g_propagate_error (error, key_file_error);
+    }
+
+  return double_value;
+}
+
+/**
+ * g_key_file_set_double:
+ * @key_file: a #GKeyFile
+ * @group_name: a group name
+ * @key: a key
+ * @value: an double value
+ *
+ * Associates a new double value with @key under @group_name.
+ * If @key cannot be found then it is created. If @group_name
+ * is %NULL, the start group is used.
+ *
+ * Since: 2.12
+ **/
+void
+g_key_file_set_double  (GKeyFile    *key_file,
+                        const gchar *group_name,
+                        const gchar *key,
+                        gdouble      value)
+{
+  gchar result[G_ASCII_DTOSTR_BUF_SIZE];
+
+  g_return_if_fail (key_file != NULL);
+  g_return_if_fail (group_name != NULL);
+  g_return_if_fail (key != NULL);
+
+  g_ascii_dtostr ( result, sizeof (result), value );
+  g_key_file_set_value (key_file, group_name, key, result);
+}
+
+/**
+ * g_key_file_get_double_list:
+ * @key_file: a #GKeyFile
+ * @group_name: a group name
+ * @key: a key
+ * @length: the number of doubles returned
+ * @error: return location for a #GError
+ *
+ * Returns the values associated with @key under @group_name as
+ * doubles. If @group_name is %NULL, the start group is used.
+ *
+ * If @key cannot be found then the return value is undefined and
+ * @error is set to #G_KEY_FILE_ERROR_KEY_NOT_FOUND. Likewise, if
+ * the values associated with @key cannot be interpreted as doubles
+ * then the return value is also undefined and @error is set to
+ * #G_KEY_FILE_ERROR_INVALID_VALUE.
+ *
+ * Return value: the values associated with the key as a double
+ *
+ * Since: 2.12
+ **/
+gdouble *
+g_key_file_get_double_list  (GKeyFile     *key_file,
+                             const gchar  *group_name,
+                             const gchar  *key,
+                             gsize        *length,
+                             GError      **error)
+{
+  GError *key_file_error = NULL;
+  gchar **values;
+  gdouble *double_values;
+  gsize i, num_doubles;
+
+  g_return_val_if_fail (key_file != NULL, NULL);
+  g_return_val_if_fail (group_name != NULL, NULL);
+  g_return_val_if_fail (key != NULL, NULL);
+
+  values = g_key_file_get_string_list (key_file, group_name, key,
+                                       &num_doubles, &key_file_error);
+
+  if (key_file_error)
+    g_propagate_error (error, key_file_error);
+
+  if (!values)
+    return NULL;
+
+  double_values = g_new0 (gdouble, num_doubles);
+
+  for (i = 0; i < num_doubles; i++)
+    {
+      double_values[i] = g_key_file_parse_value_as_double (key_file,
+							   values[i],
+							   &key_file_error);
+
+      if (key_file_error)
+        {
+          g_propagate_error (error, key_file_error);
+          g_strfreev (values);
+          g_free (double_values);
+
+          return NULL;
+        }
+    }
+  g_strfreev (values);
+
+  if (length)
+    *length = num_doubles;
+
+  return double_values;
+}
+
+/**
+ * g_key_file_set_double_list:
+ * @key_file: a #GKeyFile
+ * @group_name: a group name
+ * @key: a key
+ * @list: an array of double values
+ * @length: number of double values in @list
+ *
+ * Associates a list of double values with @key under
+ * @group_name.  If @key cannot be found then it is created.
+ * If @group_name is %NULL the start group is used.
+ *
+ * Since: 2.21
+ **/
+void
+g_key_file_set_double_list (GKeyFile     *key_file,
+			    const gchar  *group_name,
+			    const gchar  *key,
+			    gdouble       list[],
+			    gsize         length)
+{
+  GString *values;
+  gsize i;
+
+  g_return_if_fail (key_file != NULL);
+  g_return_if_fail (group_name != NULL);
+  g_return_if_fail (key != NULL);
+  g_return_if_fail (list != NULL);
+
+  values = g_string_sized_new (length * 16);
+  for (i = 0; i < length; i++)
+    {
+      gchar result[G_ASCII_DTOSTR_BUF_SIZE];
+
+      g_ascii_dtostr( result, sizeof (result), list[i] );
+
+      g_string_append (values, result);
+      g_string_append_c (values, ';');
+    }
+
+  g_key_file_set_value (key_file, group_name, key, values->str);
+  g_string_free (values, TRUE);
+}
+
 static void
 g_key_file_set_key_comment (GKeyFile             *key_file,
                             const gchar          *group_name,
@@ -2330,6 +2558,7 @@ g_key_file_get_key_comment (GKeyFile             *key_file,
                             GError              **error)
 {
   GKeyFileGroup *group;
+  GKeyFileKeyValuePair *pair;
   GList *key_node, *tmp;
   GString *string;
   gchar *comment;
@@ -2365,23 +2594,37 @@ g_key_file_get_key_comment (GKeyFile             *key_file,
    * key and concatentate them.
    */
   tmp = key_node->next;
-  while (tmp != NULL)
+  if (!key_node->next)
+    return NULL;
+
+  pair = (GKeyFileKeyValuePair *) tmp->data;
+  if (pair->key != NULL)
+    return NULL;
+
+  while (tmp->next)
     {
-      GKeyFileKeyValuePair *pair;
-
-      pair = (GKeyFileKeyValuePair *) tmp->data;
-
+      pair = (GKeyFileKeyValuePair *) tmp->next->data;
+      
       if (pair->key != NULL)
         break;
+
+      tmp = tmp->next;
+    }
+
+  while (tmp != key_node)
+    {
+      GKeyFileKeyValuePair *pair;
+      
+      pair = (GKeyFileKeyValuePair *) tmp->data;
       
       if (string == NULL)
-        string = g_string_sized_new (512);
-
+	string = g_string_sized_new (512);
+      
       comment = g_key_file_parse_value_as_comment (key_file, pair->value);
       g_string_append (string, comment);
       g_free (comment);
-
-      tmp = tmp->next;
+      
+      tmp = tmp->prev;
     }
 
   if (string != NULL)
@@ -2395,15 +2638,68 @@ g_key_file_get_key_comment (GKeyFile             *key_file,
   return comment;
 }
 
+static gchar *
+get_group_comment (GKeyFile       *key_file,
+		   GKeyFileGroup  *group,
+		   GError        **error)
+{
+  GString *string;
+  GList *tmp;
+  gchar *comment;
+
+  string = NULL;
+
+  tmp = group->key_value_pairs;
+  while (tmp)
+    {
+      GKeyFileKeyValuePair *pair;
+
+      pair = (GKeyFileKeyValuePair *) tmp->data;
+
+      if (pair->key != NULL)
+	{
+	  tmp = tmp->prev;
+	  break;
+	}
+
+      if (tmp->next == NULL)
+	break;
+
+      tmp = tmp->next;
+    }
+  
+  while (tmp != NULL)
+    {
+      GKeyFileKeyValuePair *pair;
+
+      pair = (GKeyFileKeyValuePair *) tmp->data;
+
+      if (string == NULL)
+        string = g_string_sized_new (512);
+
+      comment = g_key_file_parse_value_as_comment (key_file, pair->value);
+      g_string_append (string, comment);
+      g_free (comment);
+
+      tmp = tmp->prev;
+    }
+
+  if (string != NULL)
+    return g_string_free (string, FALSE);
+
+  return NULL;
+}
+
 static gchar *
 g_key_file_get_group_comment (GKeyFile             *key_file,
                               const gchar          *group_name,
                               GError              **error)
 {
+  GList *group_node;
   GKeyFileGroup *group;
   
-  group = g_key_file_lookup_group (key_file, group_name);
-  if (!group)
+  group_node = g_key_file_lookup_group_node (key_file, group_name);
+  if (!group_node)
     {
       g_set_error (error, G_KEY_FILE_ERROR,
                    G_KEY_FILE_ERROR_GROUP_NOT_FOUND,
@@ -2413,20 +2709,21 @@ g_key_file_get_group_comment (GKeyFile             *key_file,
       return NULL;
     }
 
+  group = (GKeyFileGroup *)group_node->data;
   if (group->comment)
     return g_strdup (group->comment->value);
-
-  return NULL;
+  
+  group_node = group_node->next;
+  group = (GKeyFileGroup *)group_node->data;  
+  return get_group_comment (key_file, group, error);
 }
 
 static gchar *
 g_key_file_get_top_comment (GKeyFile             *key_file,
                             GError              **error)
 {
-  GList *group_node, *tmp;
+  GList *group_node;
   GKeyFileGroup *group;
-  GString *string;
-  gchar *comment;
 
   /* The last group in the list should be the top (comments only)
    * group in the file
@@ -2436,40 +2733,7 @@ g_key_file_get_top_comment (GKeyFile             *key_file,
   group = (GKeyFileGroup *) group_node->data;
   g_assert (group->name == NULL);
 
-  string = NULL;
-
-  /* Then find all the comments already associated with the
-   * key and concatentate them.
-   */
-  tmp = group->key_value_pairs;
-  while (tmp != NULL)
-    {
-      GKeyFileKeyValuePair *pair;
-
-      pair = (GKeyFileKeyValuePair *) tmp->data;
-
-      if (pair->key != NULL)
-        break;
-      
-      if (string == NULL)
-        string = g_string_sized_new (512);
-
-      comment = g_key_file_parse_value_as_comment (key_file, pair->value);
-      g_string_append (string, comment);
-      g_free (comment);
-
-      tmp = tmp->next;
-    }
-
-  if (string != NULL)
-    {
-      comment = string->str;
-      g_string_free (string, FALSE);
-    }
-  else
-    comment = NULL;
-
-  return comment;
+  return get_group_comment (key_file, group, error);
 }
 
 /**
@@ -2479,14 +2743,15 @@ g_key_file_get_top_comment (GKeyFile             *key_file,
  * @key: a key
  * @error: return location for a #GError
  *
- * Retreives a comment above @key from @group_name.
+ * Retrieves a comment above @key from @group_name.
  * @group_name. If @key is %NULL then @comment will
  * be read from above @group_name.  If both @key
  * and @group_name are NULL, then @comment will
  * be read from above the first group in the file.
  *
- * Since: 2.6
  * Returns: a comment that should be freed with g_free()
+ *
+ * Since: 2.6
  **/
 gchar * 
 g_key_file_get_comment (GKeyFile             *key_file,
@@ -2610,7 +2875,9 @@ g_key_file_add_group (GKeyFile    *key_file,
 
   g_return_if_fail (key_file != NULL);
   g_return_if_fail (group_name != NULL);
-  g_return_if_fail (g_key_file_lookup_group_node (key_file, group_name) == NULL);
+
+  if (g_key_file_lookup_group_node (key_file, group_name) != NULL)
+    return;
 
   group = g_new0 (GKeyFileGroup, 1);
   group->name = g_strdup (group_name);
@@ -2852,15 +3119,12 @@ g_key_file_lookup_group_node (GKeyFile    *key_file,
   GKeyFileGroup *group;
   GList *tmp;
 
-  group = NULL;
   for (tmp = key_file->groups; tmp != NULL; tmp = tmp->next)
     {
       group = (GKeyFileGroup *) tmp->data;
 
       if (group && group->name && strcmp (group->name, group_name) == 0)
         break;
-
-      group = NULL;
     }
 
   return tmp;
@@ -2934,14 +3198,9 @@ g_key_file_line_is_group (const gchar *line)
 
   p = g_utf8_next_char (p);
 
-  if (!*p)
-    return FALSE;
-
-  p = g_utf8_next_char (p);
-
   /* Group name must be non-empty
    */
-  if (*p == ']')
+  if (!*p || *p == ']')
     return FALSE;
 
   while (*p && *p != ']')
@@ -3202,6 +3461,24 @@ g_key_file_parse_integer_as_value (GKeyFile *key_file,
   return g_strdup_printf ("%d", value);
 }
 
+static gdouble
+g_key_file_parse_value_as_double  (GKeyFile     *key_file,
+                                   const gchar  *value,
+                                   GError      **error)
+{
+  gchar *end_of_valid_d;
+  gdouble double_value = 0;
+
+  double_value = g_ascii_strtod (value, &end_of_valid_d);
+
+  if (*end_of_valid_d != '\0')
+    g_set_error (error, G_KEY_FILE_ERROR,
+                 G_KEY_FILE_ERROR_INVALID_VALUE,
+                 _("Value '%s' cannot be interpreted as a float number."), value);
+
+  return double_value;
+}
+
 static gboolean
 g_key_file_parse_value_as_boolean (GKeyFile     *key_file,
 				   const gchar  *value,
