@@ -50,6 +50,7 @@ typedef struct
     gint integer;
     gchar *str;
     gchar **array;
+    gdouble dbl;
   } prev;
   union 
   {
@@ -117,36 +118,31 @@ static void free_pending_nulls (GOptionContext *context,
 GQuark
 g_option_error_quark (void)
 {
-  static GQuark q = 0;
-  
-  if (q == 0)
-    q = g_quark_from_static_string ("g-option-context-error-quark");
-
-  return q;
+  return g_quark_from_static_string ("g-option-context-error-quark");
 }
 
 /**
  * g_option_context_new:
  * @parameter_string: a string which is displayed in
- *    the first line of <option>--help</option> output, after the 
+ *    the first line of <option>--help</option> output, after the
  *    usage summary 
- *    <literal><replaceable>programname</replaceable> [OPTION...]</literal>.
+ *    <literal><replaceable>programname</replaceable> [OPTION...]</literal>
  *
  * Creates a new option context. 
  *
  * The @parameter_text can serve multiple purposes. It can be used
- * to add descriptions for "rest" arguments, which are not parsed by 
- * the #GOptionContext, typically something like "FILES" or 
- * "FILE1 FILE2...". (If you are using #G_OPTION_REMAINING for 
- * collecting "rest" arguments, GLib handles this automatically by 
- * using the @arg_description of the corresponding #GOptionEntry in 
+ * to add descriptions for "rest" arguments, which are not parsed by
+ * the #GOptionContext, typically something like "FILES" or
+ * "FILE1 FILE2...". (If you are using #G_OPTION_REMAINING for
+ * collecting "rest" arguments, GLib handles this automatically by
+ * using the @arg_description of the corresponding #GOptionEntry in
  * the usage summary.)
  *
  * Another common usage is to give a summary of the program
- * functionality. This can be a short summary on the same line, 
- * like " - frob the strings", or a longer description in a paragraph 
- * below the usage summary. In this case, @parameter_string should start 
- * with two newlines, to separate the description from the usage summary:
+ * functionality. This can be a short summary on the same line,  
+ * like " - frob the strings", or a longer description in a paragraph  
+ * below the usage summary. In this case, @parameter_string should start  
+ * with two newlines, to separate the description from the usage summary:  
  * "\n\nA program to frob strings, which will..."
  *
  * Returns: a newly created #GOptionContext, which must be
@@ -684,6 +680,42 @@ parse_int (const gchar *arg_name,
   return TRUE;
 }
 
+
+static gboolean
+parse_double (const gchar *arg_name,
+	   const gchar *arg,
+	   gdouble        *result,
+	   GError     **error)
+{
+  gchar *end;
+  gdouble tmp;
+
+  errno = 0;
+  tmp = g_strtod (arg, &end);
+  
+  if (*arg == '\0' || *end != '\0')
+    {
+      g_set_error (error,
+		   G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+		   _("Cannot parse double value '%s' for %s"),
+		   arg, arg_name);
+      return FALSE;
+    }
+  if (errno == ERANGE)
+    {
+      g_set_error (error,
+		   G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+		   _("Double value '%s' for %s out of range"),
+		   arg, arg_name);
+      return FALSE;
+    }
+
+  *result = tmp;
+  
+  return TRUE;
+}
+
+
 static Change *
 get_change (GOptionContext *context,
 	    GOptionArg      arg_type,
@@ -735,7 +767,9 @@ parse_arg (GOptionContext *context,
      
 {
   Change *change;
-  
+
+  g_assert (value || OPTIONAL_ARG (entry) || NO_ARG (entry));
+
   switch (entry->arg)
     {
     case G_OPTION_ARG_NONE:
@@ -900,6 +934,23 @@ parse_arg (GOptionContext *context,
 	
 	return retval;
 	
+	break;
+      }
+    case G_OPTION_ARG_DOUBLE:
+      {
+	gdouble data;
+
+	if (!parse_double (option_name, value,
+ 			&data,
+			error))
+	  {
+	    return FALSE;
+	  }
+
+	change = get_change (context, G_OPTION_ARG_DOUBLE,
+			     entry->arg_data);
+	change->prev.dbl = *(gdouble *)entry->arg_data;
+	*(gdouble *)entry->arg_data = data;
 	break;
       }
     default:
@@ -1166,6 +1217,9 @@ free_changes_list (GOptionContext *context,
 	      g_strfreev (change->allocated.array.data);
 	      *(gchar ***)change->arg_data = change->prev.array;
 	      break;
+	    case G_OPTION_ARG_DOUBLE:
+	      *(gdouble *)change->arg_data = change->prev.dbl;
+	      break;
 	    default:
 	      g_assert_not_reached ();
 	    }
